#!/usr/bin/env python
# -*- coding: utf-8 -*-
#       input.py
#
#       Copyright 2008 Gordon Allott <gordallott@gmail.com>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
#
#       #

import os, sys, math, random
import pygame
from pygame.locals import *



class Input:

    key_list = {}
    quit_callback = None


    def add_key(self, key_sym, callback, repeating=False):
        """ adds a key callback for the specified sym and callback """

        type = 'pressed_cbs' if not repeating else 'repeating_cbs'
        store = self._get_key(key_sym)[type]

        if callback in store:
            return

        store.append(callback)

    def remove_key(self, key_sym, callback, repeating=False):
        type = 'pressed_cbs' if not repeating else 'repeating_cbs'
        store = self._get_key(key_sym)[type]

        if not callback in store:
            return

        else:
            store.pop(store.index(callback))


    def add_quit(self, callback):
        """ on quit goes here """
        self.quit_callback = callback

    def step(self):
        """ handles all events, fires off callbacks """

        events = pygame.event.get()

        for event in events:
            if event.type == QUIT:
                if self.quit_callback:
                    self.quit_callback()
                    continue

            if event.type == KEYDOWN:
                # first of all we need to turn the key ispressed for this key
                # on (for repeating callbacks)


                key = self._get_key(event.key)
                key['is_pressed'] = True

                #check for on pressed callbacks
                for callback in key['pressed_cbs']:
                    callback(event.key)

            if event.type == KEYUP:
                key = self._get_key(event.key)
                key['is_pressed'] = False # turn off repeating keys


        self.emit_repeating()



    def emit_repeating(self):
        """ goes through all our keys looking for repeating callbacks """

        for keyname, key in self.key_list.iteritems():
            if key['is_pressed']:
                for callback in key['repeating_cbs']:
                    callback(keyname)


    def _get_key(self, key_sym):
        if self.key_list.has_key(key_sym):
            return self.key_list[key_sym]

        # our key doesn't exist so we create a 'base' key
        key_proto = {
            'is_pressed': False,
            'pressed_cbs': [],
            'repeating_cbs': [],
        }

        self.key_list[key_sym] = key_proto

        return self.key_list[key_sym]

