'''Game main module.

Contains the entry point used by the run_game.py script.

Feel free to put all your game code here, or in other modules in this "gamelib"
package.
'''

import OpenGL
OpenGL.ERROR_CHECKING = False
OpenGL.ERROR_ON_COPY = True
import os, sys, math, random
from numpy import *
import pygame
from pygame.locals import *
import ode
import random
from math import *
from OpenGL.GL import *
import cProfile

import data
import input
import render
import audio
import world.gameworld
try:
    import shader
except:
    print 'problem importing shader code, disabling shaders'

#from scenes import intro
#from scenes import menu
from gamelib.scenes import level_one

from gamelib.scenes import level_zero
#from scenes import final

import instance

class MainLoop:

    scenes = [
        level_zero.level_zero,
        level_one.level_one,
    ]

    current_scene_index = 0
    current_scene = None

    quit = False

    def __init__(self, fullscreen=False, do_shaders=True, screen_size=(800, 600), level=0):
        """ init stuff go here """

        instance.instance = self

        self.current_scene_index = level


        self.render = render.RenderLib(size=screen_size, fullscreen=fullscreen)
        self.clock = pygame.time.Clock()

        self.audio = audio.AudioSystem()

        pygame.joystick.init()

        self.input = input.Input()
        self.input.add_quit(self.cb_quit)
        self.input.add_key(K_ESCAPE, self.cb_quit, repeating=False)
        self.current_scene = self.scenes[self.current_scene_index](self.render, self.input)

        if do_shaders:
            try:

                self.screen_FB = shader.Framebuffer(self.render, screen_size, smooth=False, backbuffer=True)
                self.Bloom = shader.PostProcess_Bloom(self.render)
                self.Blur = shader.PostProcess_Blur(self.render)
                self.shader_activate = True
            except:
                print "Ohoh... you don't have the capabilities to run our shaders :( - i'm gonna disable them so you can still play"
                self.shader_activate = False

        else:
            self.shader_activate = False




    def cb_quit(self, *args):
        self.quit = True

    def run(self):
        while not self.quit:
            self.input.step()
            self.audio.step()

            ##do our shader stuff
            if self.shader_activate:
                self.render.reset()
                self.render.set_screen_pos(0.0, 0.0)
                self.render.set_screen_rot(0.0)
                self.render.set_screen_zoom(1.0)

                self.screen_FB.bind()

            self.render.set_projection()

            self.step()

            self.render.reset()


            if self.shader_activate:
                self.render.set_screen_pos(0.0, 0.0)
                self.render.set_screen_rot(0.0)
                self.render.set_screen_zoom(1.0)
                self.render.disable(GL_BLEND)
                self.render.set_ambiant(1.0, 1.0, 1.0)
                self.render.set_color(1.0, 1.0, 1.0, 1.0)
                self.render.set_screen()
                self.render.set_position(0.0, 0.0, 0.0)
                self.render.scale(1.0, 1.0)

                self.screen_FB.unbind()
                self.screen_FB.test_fb()

                pp_tex = self.Blur.process(self.screen_FB.get_texture())

                if True:
                    self.Bloom.process(self.screen_FB.get_texture(), pp_tex) # do our post-process bloom (using the blur texture)
                else:
                    self.Bloom.process(pp_tex, pp_tex)

            self.render.end_frame(do_clear=False)
            self.clock.tick(30)

    def step(self):
        """ do a step """

        scene = self.current_scene

        scene.step()

        if scene.stop_scene:
            scene.shutdown()
            del(scene)
            self.current_scene_index += 1
            self.current_scene = self.scenes[self.current_scene_index](self.render, self.input)





def usage():
    print """woo :) - options:
--fullscreen - enables fullscreen mode
--noshaders - for people with a weedy gpu or have problems with shaders
"""

def main(*args):

    try:
        import psyco
        psyco.full()
    except ImportError:
        pass

    import getopt


    try:
        opts, args = getopt.getopt(sys.argv[1:], "hsf", ["help", "noshaders", "fullscreen", "resolution=", "widescreen", "level="])
    except getopt.GetoptError, err:
        # print help information and exit:
        print str(err) # will print something like "option -a not recognized"
        usage()
        sys.exit(2)

    fullscreen = False
    shaders = True
    resolution = (800, 600)
    level = 0

    for o, a in opts:
        if o in ("-s", "--noshaders"):
            shaders = False

        elif o in ("-h", "--help"):
            usage()
            sys.exit()

        elif o in ("-f", "--fullscreen"):
            fullscreen = True

        elif o in ("--widescreen"):
            resolution = (800, 450)

        elif o in ("--resolution"):
            try:
                resolution = a.split('x')
                resolution = (int(resolution[0]), int(resolution[1]))
            except:
                print 'sorry,', a, 'is not a valid string, resolution should be in the form of --resolution=800x600 or something'
                sys.exit()

        elif o in ("--level"):
            level = int(a)

        else:
            assert False, "unhandled option"


    mainloop = MainLoop(fullscreen=fullscreen, do_shaders=shaders, screen_size=resolution, level=level)
    mainloop.run()
