#!/usr/bin/env python
# -*- coding: utf-8 -*-
#       render.py
#
#       Copyright 2008 Gordon Allott <gordallott@gmail.com>
#
#       This program is free software; you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation; either version 3 of the License, or
#       (at your option) any later version.
#
#       This program is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with this program; if not, write to the Free Software
#       Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#       MA 02110-1301, USA.
#
#       #
import os, sys, math, random
from numpy import *
import pygame
from pygame.locals import *
import ode
import random
import re
from math import *
from OpenGL.GL import *
from OpenGL.GLU import *
from OpenGL.GLUT import *
from OpenGL.GL.EXT.framebuffer_object import *
from OpenGL.GL.ARB.multitexture import *

import data as Data

Infinity = 10000.0

class RenderLib:

    enables = []
    tex = 0

    texture0 = None
    texture1 = None
    texture2 = None
    texture3 = None

    current_texture = 0

    blend_func = None
    base_scalar = (1.0, 1.0, 1.0)
    base_rotation = (0.0, 0.0, 0.0)
    base_position = (0.0, 0.0, -1.0)
    position = (0.0, 0.0, -0.0)
    rotation = (0.0, 0.0, 0.0)
    scalar = (1.0, 1.0)

    base_color_scale = (1.0, 1.0, 1.0, 1.0)
    color = (1.0, 1.0, 1.0, 1.0)

    def __init__(self, size=(800, 600), fullscreen=False):

        self.size = size
        self.fullscreen = fullscreen

        self.clock = pygame.time.Clock()
        self.time = 0.0

        self.create_screen()
        self.create_context()

    def create_screen(self):
        pygame.display.init()


        pygame.display.set_mode(self.size, OPENGL|DOUBLEBUF)

        if self.fullscreen:
            pygame.display.toggle_fullscreen()


    def get_screen_bounds(self):
        width = (self.swidth) / self.base_scalar[0]
        height = (self.sheight) / self.base_scalar[1]

        box = max(width, height) * 0.5
        x, y, z = self.base_position
        x, y = (-x, -y)

        return ((x - box, x + box), (y - box, y + box))

        # we just return width, width as width will always

    def create_context(self):
        glClearColor(0.0, 0.0, 0.0, 0.0)
        self.set_projection()


    def set_projection(self):
        width, height = self.size
        glViewport(0, 0, int(width), int(height))

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        self.aspect = self.size[0] / float(self.size[1])

        self.swidth = 10.0 * self.aspect
        self.sheight = 10.0

        glOrtho(-(self.swidth*0.5), self.swidth*0.5,
                -(self.sheight*0.5), self.sheight*0.5,
                -22.0, 22.0)


        glMatrixMode(GL_MODELVIEW)

        glLoadIdentity()

    def end_frame(self, do_clear=False):
        """ ends the current gl frame """
        pygame.display.flip()

        self.set_projection()

        if (do_clear):
            glClear(GL_COLOR_BUFFER_BIT)

        self.clock.tick()
        ticks = self.clock.get_time()
        self.time += ticks / 1000.0

    def clearscreen(self):
        glClear(GL_COLOR_BUFFER_BIT)

    def reset_rotation(self):
        """ resets the rotation to 0 """
        glRotate(self.rotation[2],  0.0,  0.0, -1.0)
        glRotate(self.rotation[1],  0.0, -1.0,  0.0)
        glRotate(self.rotation[0], -1.0,  0.0,  0.0)

        self.rotation = (0.0, 0.0, 0.0)

    def reset_scale(self):
        """ resets the scale to 1.0 """
        glScale(1.0/self.scalar[0], 1.0/self.scalar[1], 1.0)
        self.scalar = (1.0, 1.0)

    def reset_position(self):
        """ resets the translation to 0 """
        glTranslate(-self.position[0], -self.position[1], -self.position[2])

        self.position = (0.0, 0.0, 0.0)

    def reset(self):
        self.reset_scale()
        self.reset_rotation()
        self.reset_position()

        glLoadIdentity()

    def rotate(self, x, y, z):
        """ rotate to the specified degrees """

        cx, cy, cz = self.rotation

        glRotate(x - cx, 1.0, 0.0, 0.0)
        glRotate(y - cy, 0.0, 1.0, 0.0)
        glRotate(z - cz, 0.0, 0.0, 1.0)

        self.rotation = (x, y, z)

    def scale(self, x, y):
        """ scale to the specified scalar """
        x = float(x)
        y = float(y)

        self.reset_scale()

        glScale(x, y, 1.0)

        self.scalar = (x, y)

    def set_position(self, x, y, z=-1):
        """ translate to the specified position """
        self.reset_scale()
        self.reset_rotation()

        cx, cy, cz = self.position

        glTranslate(x - cx, y - cy, z - cz)

        self.position = (x, y, z)

    def set_screen_zoom(self, zoom):
        self.base_scalar = (zoom, zoom, 1.0)

    def set_screen_pos(self, x, y):
        self.base_position = (x, y, 0.0)

    def set_screen_rot(self, angle):
        self.base_rotation = (0.0, 0.0, angle)

    def set_ambiant(self, *args):
        self.base_color_scale = (args[0], args[1], args[2], 1.0)

    def set_color(self, *args):
        if type(args) == type(()):
            args = list(args)

        if len(args) < 4:
            args.append(1.0)

        self.color = (self.base_color_scale[0] * args[0],
                      self.base_color_scale[1] * args[1],
                      self.base_color_scale[2] * args[2],
                      self.base_color_scale[3] * args[3])

        glColor4fv(self.color)



    def set_screen(self):
        """ should be called once per frame to set the screen zoom/pos/rot"""
        glScale(self.base_scalar[0], self.base_scalar[1], self.base_scalar[2])
        glRotate(self.base_rotation[2], 0.0, 0.0, 1.0)
        glTranslate(self.base_position[0], self.base_position[1], self.base_position[2])




    def enable(self, token):

        if token in self.enables:
            return

        self.enables.append(token)
        glEnable(token)

    def disable(self, token):

        if token in self.enables:
            self.enables.pop(self.enables.index(token))

        glDisable(token)

    def set_blend(self, type):
        """ sets the required blend function

        @param type: a blend type, may be 'replace', 'addictive', 'subtractive', 'modulate'
        """

        switch = {
            'replace': (0, 0),
            'addictive': (GL_ONE, GL_ONE),
            'subtractive': (GL_ZERO, GL_ONE_MINUS_SRC_COLOR),
            'modulate': (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA),
        }

        mode = switch[type]

        if type == 'replace':
            self.disable(GL_BLEND)
        else:
            self.enable(GL_BLEND)

        if self.blend_func == mode:
            return

        glBlendFunc(mode[0], mode[1])
        self.blend_func = mode

    def select_texture(self, unit):

        if self.current_texture == unit:
            return

        if unit == 0:
            glActiveTexture(GL_TEXTURE0)

        if unit == 1:
            glActiveTexture(GL_TEXTURE1)

        if unit == 2:
            glActiveTexture(GL_TEXTURE2)

        if unit == 3:
            glActiveTexture(GL_TEXTURE3)

        if unit > 3 or unit < 0:
            return

        self.current_texture = unit

    def unbind_textures(self):

        if self.texture3:
            self.select_texture(3)
            glBindTexture(GL_TEXTURE_2D, 0);
            self.texture3 = None

        if self.texture2:
            self.select_texture(2)
            glBindTexture(GL_TEXTURE_2D, 0);
            self.texture2 = None

        if self.texture1:
            self.select_texture(1)
            glBindTexture(GL_TEXTURE_2D, 0);
            self.texture1 = None

        if self.texture0:
            self.select_texture(0)
            glBindTexture(GL_TEXTURE_2D, 0);
            self.texture0 = None

    def bind_texture(self, *textures):

        self.unbind_textures()

        for texnum in reversed(range(0, len(textures))):
            self.select_texture(texnum)
            glBindTexture(GL_TEXTURE_2D, textures[texnum].get_id())

            if texnum == 0:
                self.texture0 = textures[texnum]

            if texnum == 1:
                self.texture1 = textures[texnum]

            if texnum == 2:
                self.texture2 = textures[texnum]

            if texnum == 3:
                self.texture3 = textures[texnum]

    textures = {}

    def create_texture(self, filename, **kwargs):

        filepath = Data.filepath(filename)
        if not os.path.exists(filepath):
            raise Exception, 'path %s does not exist' % (filepath)

        if self.textures.has_key(filepath):
            texture = self.textures[filepath]
        else:

            if os.path.isdir(filepath):
                texture = Animation(self, filename, **kwargs)
            else:
                if kwargs.has_key('repeat'):
                    kwargs.pop('repeat')
                if kwargs.has_key('fps'):
                    kwargs.pop('fps')
                texture = Texture(self, filename=filename, **kwargs)

            self.textures[filepath] = texture

        return texture


class DisplayList:

    id = None

    def __init__(self):
        self.id = glGenLists(1)

    def compile(self):
        glNewList(self.id, GL_COMPILE)

    def end_compile(self):
        glEndList();

    def execute(self):
        glCallList(self.id)


class Animation:
    render = None
    filepath = None
    textures = []
    repeat = True

    current_frame = 0

    last_framechange = 0
    num_iterations = 0

    def __init__(self, renderobj, filepath, fps=12, repeat=True,
                 smooth=True, mipmap=True, repeats=('repeat', 'repeat')):

        self.render = renderobj
        self.fps = fps
        self.filepath = Data.filepath(filepath)
        self.repeat = repeat

        # first we go through the directory given and scan for png files

        listfiles = os.listdir(self.filepath)
        files = []
        for file in listfiles:
            if file[-3:] == 'png':
                files.append(file)

        # we now have a list of png files in a directory, whoop?
        def order_list(x, y):
            x = x.lower()
            y = y.lower()

            cmp = r'.*(?P<major>\d+)_(?P<minor>\d+).png'

            xinfo = re.match(cmp, x).groupdict()
            yinfo = re.match(cmp, y).groupdict()

            if int(xinfo['major']) != int(yinfo['major']):
                return int(xinfo['major']) - int(yinfo['major'])

            return int(xinfo['minor']) - int(yinfo['minor'])


        files.sort(order_list)

        # we now have a list of png files in the *correct* order designed to
        # work with gimps 'slice' function

        self.textures = []
        for file in files:
            file = os.path.join(filepath, file)
            texture = self.render.create_texture(file, smooth=smooth, mipmap=mipmap,
                                                 repeats=repeats)

            self.textures.append(texture)


        self.last_framechange = self.render.time
        self.size = self.textures[0].size
        self.aspect = self.size[1] / float(self.size[0])



    def get_id(self):
        #check time
        needed_time = (1.0 / self.fps)

        if self.render.time - self.last_framechange >= needed_time:
            self.current_frame += 1

            if self.current_frame == len(self.textures) and not self.repeat:
                self.current_frame -= 1

            self.current_frame = self.current_frame % len(self.textures)

            self.last_framechange = self.render.time

        return self.textures[self.current_frame].get_id()



class Texture:

    def __init__(self, renderlib, filename=None, data=None, id=None, size=(1, 1),
                 smooth=True, mipmap=True, repeats=('repeat', 'repeat')):

        self.render = renderlib

        width, height = (None, None)
        self.repeats = (GL_REPEAT, GL_REPEAT)

        if repeats[0] == 'clamp':
            self.repeats = (GL_CLAMP, self.repeats[1])

        if repeats[1] == 'clamp':
            self.repeats = (self.repeats[0], GL_CLAMP)


        if filename:
            filepath = Data.filepath(filename)
            surf = pygame.image.load(filepath)
            data = pygame.image.tostring(surf, "RGBA", 1)
            width = surf.get_width()
            height = surf.get_height()

        elif data:
            data, width, height = data, size[0], size[1]


        elif id:
            id, width, height = id

        else:
            raise Exception, 'bad texture data'


        if data:
            texture = glGenTextures(1)
            glBindTexture(GL_TEXTURE_2D, texture)

            if not mipmap:
                glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0,
                             GL_RGBA, GL_UNSIGNED_BYTE, data )

            else:
                gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, width, height,
                                  GL_RGBA, GL_UNSIGNED_BYTE, data)

            id = texture


        glBindTexture(GL_TEXTURE_2D, id)

        # here we now *definatly* have id, width, height

        self.id = id
        self.size = (width, height)
        self.aspect = self.size[1] / float(self.size[0])

        self.filter_min = GL_LINEAR
        self.filter_max = GL_LINEAR

        if not smooth:
            self.filter_min, self.filter_max = (GL_NEAREST, GL_NEAREST)

        if mipmap:
            self.filter_min = GL_LINEAR_MIPMAP_LINEAR

        self.set_tex_params()


    def set_tex_params(self):

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, self.filter_min)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, self.filter_max)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, self.repeats[0])
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, self.repeats[1])

    def get_id(self):
        return self.id


class Sprite:

    position = (0.0, 0.0)
    flip = (False, False)
    color = (1.0, 1.0, 1.0, 1.0)

    def __init__(self, renderlib, texture, size=(1.0, 1.0), tile_repeats=(1.0, 1.0),
                 blend='modulate'):
        self.render = renderlib
        self.texture = texture
        self.display_list = DisplayList()
        self.size = size
        self.blend = blend
        self.flip = (False, False)
        self.color = (1.0, 1.0, 1.0, 1.0)
        self.rotate = None

        ## build our quad

        self.display_list.compile()

        width = size[0] * 0.5
        height = size[1] * 0.5

        glBegin(GL_QUADS)

        glTexCoord2f(0.0, 0.0); glVertex3f(-width, -height, 0.0)
        glTexCoord2f(0.0, 1.0 * tile_repeats[1]); glVertex3f(-width,  height, 0.0)
        glTexCoord2f(1.0 * tile_repeats[0], 1.0 * tile_repeats[1]); glVertex3f( width,  height, 0.0)
        glTexCoord2f(1.0 * tile_repeats[0], 0.0); glVertex3f( width, -height, 0.0)

        glEnd()

        self.display_list.end_compile()


    def check_bounds(self):
        position = self.position
        size = max(self.size) * 0.5
        (left, right), (bottom, top) = self.render.get_screen_bounds()

        # check left
        out_of_bounds_x, out_of_bounds_y = (True, True)

        if position[0] + size > left and position[0] - size < right:
            out_of_bounds_x = False

        if position[1] + size > bottom and position[1] - size < top:
            out_of_bounds_y = False

        if not out_of_bounds_x and not out_of_bounds_y:
            return False

        return True


    def draw(self, mat=None, bounds_check=True, colors=True):

        if bounds_check:
            if self.check_bounds():
                return


        if colors:
            self.render.set_color(*self.color)

        if self.blend:
            self.render.enable(GL_BLEND)
            self.render.set_blend(self.blend)

        self.render.enable(GL_TEXTURE_2D)
        self.render.bind_texture(self.texture)
        self.render.set_position(self.position[0], self.position[1])

        if self.flip == (True, True):
            self.render.scale(-1.0, -1.0)

        elif self.flip == (True, False):
            self.render.scale(-1.0, 1.0)

        elif self.flip == (False, True):
            self.render.scale(1.0, -1.0)

        else:
            self.render.scale(1.0, 1.0)
        # do set position

        if self.rotate:
            self.render.rotate(*self.rotate)
        else:
            self.render.rotate(0.0, 0.0, 0.0)

        if mat:
            glPushMatrix()
            glMultMatrixd(mat)

        self.display_list.execute()

        if mat:
            glPopMatrix()


